"""
Shared constants for the IDTAP library.
"""

# Fundamental frequency validation constants
MIN_FUNDAMENTAL_HZ = 50   # Minimum reasonable fundamental frequency for musical analysis
MAX_FUNDAMENTAL_HZ = 400  # Maximum reasonable fundamental frequency for musical analysis